%%% -*- coding: latin-1 -*-

%%% The  contents of this  file are  subject to  the Erlang  Public License,
%%% Version  1.1, (the  "License");  you may  not  use this  file except  in
%%% compliance  with the License.  You should  have received  a copy  of the
%%% Erlang  Public License  along  with this  software.  If not,  it can  be
%%% retrieved at http://plc.inf.elte.hu/erlang/
%%%
%%% Software  distributed under  the License  is distributed  on an  "AS IS"
%%% basis, WITHOUT  WARRANTY OF ANY  KIND, either expressed or  implied. See
%%% the License  for the specific language governing  rights and limitations
%%% under the License.
%%%
%%% The Original Code is RefactorErl.
%%%
%%% The Initial Developer of the  Original Code is Etvs Lornd University.
%%% Portions created  by Etvs  Lornd University are  Copyright 2008-2009,
%%% Etvs Lornd University. All Rights Reserved.

%%% ============================================================================
%%% Module information

%%% @doc

%%% @author Zsofia Arvay <zsofia.arvay@gmail.com>

-module(refusr_layer).
-vsn("$Rev: 6694 $").

-export([check_layered_architecture/2, show_insults/2]).

-include("user.hrl").

%% @doc Inputs of the function are two lists. The elements of the first list 
%% define architecture layers as group of modules. Every group of modules must 
%% have a name. At default functions in a group can call functions from the 
%% same group or from the group directly under own group. But the 
%% second list defines additional permits, allowing function 
%% calls between two architecture layers given by it's names. After this an 
%% input can look like this: [{l1,[mod1,mod2]},{l2,[mod3,mod4,mod5]},{l3,[mod6]}], 
%% [{l1,l2}]. The function checks wether in an architecture defined by the 
%% input are functions, that insult the hierarchy or not. The output is 
%% an empty list if function calls are allowed in the defined architecture 
%% or list of disabled function calls in the following format: 
%% [{{caller fun, caller fun's module},{called fun, called fun's module}}, ..].
%% In the output functions and modules are given by nodes, but in the input 
%% we can define modules by it names, too.
check_layered_architecture(Mods=[{_,[Elem|_]}|_],ArchList)->
	if
		is_tuple(Elem) -> ModList = Mods;
		is_atom(Elem) -> 
					ModList = 
						[{Cimke,[?Query:exec(?Mod:find(Mod)) || Mod<-List_E]} 
						|| {Cimke,List_E}<-Mods]
	end,
	FunList = 
		[{Cimke,?Query:exec(Mod, ?Mod:locals())} 
		|| {Cimke,Mod}<-ModList],
	FunCalls = 
		[{Cimke,[?Query:exec(Fun, ?Fun:funcalls()) || Fun<-FunListE]} 
		|| {Cimke,FunListE}<-FunList],
	FunCallPair = 
		[{Cimke,lists:zip(FunListE,FunCallsE)} 
		|| {{Cimke,FunListE},{_,FunCallsE}}<-lists:zip(FunList,FunCalls)],
	FunsList = lists:flatten([FunListE || {_,FunListE}<-FunList]),
	FunListSub = lists:sublist(FunList, length(FunList)-1),
	[DisabledFunListFirst|DisabledFunListTail] = 
		[{Cimke,FunsList--Element} || {Cimke,Element}<-FunList],
	DisabledFunctionsListTmp = 
		[DisabledFunListFirst | 
		[{Cimke,Element2--Element1} 
		|| {{_,Element1},{Cimke,Element2}}
			<-lists:zip(FunListSub,DisabledFunListTail)]],

	DisabledFunctionsList = 
		process_archlist(FunList,DisabledFunctionsListTmp,ArchList),

	_Result = 
		lists:flatten([check_disabled_functions(Element1,Element2) 
		|| {{_,Element1},{_,Element2}}
			<-lists:zip(FunCallPair,DisabledFunctionsList)]).

%% Iterates through the elements of the Archlist. Archlist contains tuples, 
%% that contain two names of module groups and means that function calls from the
%% first module group given by the fisrt element of the tuple allowed to the 
%% second module group given by the second element. If does not exists any 
%% module group under the name ignores it, but a warning will be written to the 
%% output. In each step the function deletes from the list of the disabled
%% functions the allowed functions defined by the tuple.
process_archlist(_,DisabledFunctionsList,[])->
	DisabledFunctionsList;
process_archlist(FunList,DisabledFunctionsList,[{From,To}|Tail])->
	Ret1 = lists:keyfind(To,1,FunList),
	if
		is_tuple(Ret1) -> {_,FunsList} = Ret1,
		Ret2 = lists:keyfind(From,1,DisabledFunctionsList),
		if
			is_tuple(Ret2) -> {Key,DisabledFunsList} = Ret2,
			NewDisabledFunctionsList = 
				lists:keyreplace(Key,1,DisabledFunctionsList,
								{Key,DisabledFunsList--FunsList}),
			process_archlist(FunList,NewDisabledFunctionsList,Tail);
			true -> io:format("Warning: Label named ~p not exsist.\n", [From]),
				process_archlist(FunList,DisabledFunctionsList,Tail)
		end;
		true -> io:format("Warning: Label named ~p not exsist.\n", [To]),
			process_archlist(FunList,DisabledFunctionsList,Tail)			
	end.

show({{Afirst,Asecond},{Bfirst,Bsecond}})->
	io_lib:format("{~p:~p/~p,~p:~p/~p}",
		[?Mod:name(Afirst),?Fun:name(Asecond),
		?Fun:arity(Asecond),?Mod:name(Bfirst),
		?Fun:name(Bsecond),?Fun:arity(Bsecond)]).

%% @doc The input is the same as the input of the check_layered_architecture 
%% function. The function takes the result of that function and writes it to 
%% the output in a more readable format: 
%% {error|[{caller fun's mod name:caller fun's name/caller fun's arity,
%% called fun's module name:called fun's name/called fun's arity}, ..]}.
show_insults(Mods,ArchList)->
	Result = check_layered_architecture(Mods,ArchList),
	if
		Result == [] -> [];
		true -> "{error|["++[io_lib:format("~n",[])++show(Element) || 
							Element<-Result]++"]}"
	end.

make_pair(_,[])->[];
make_pair(Func,[Head|Tail])->
	[{{lists:nth(1,?Query:exec(Func,?Fun:module())),Func},
	{lists:nth(1,?Query:exec(Head,?Fun:module())),Head}}]
	++make_pair(Func,Tail).

%% Checks wether a function is not in the list of disabled functions.
%% Returns with the list of the disabled function pairs like this:
%% [{caller, caller's module}, {called fun, called fun's module}, ..].	
check_disabled_functions([],_) ->[];
check_disabled_functions(_,[]) ->[];
check_disabled_functions([{_,[]}|TailA],B) ->
	check_disabled_functions(TailA,B);
check_disabled_functions([{Func,Lista}|TailA],B) ->
	case lists_intersection(lists:sort(Lista),lists:sort(B)) of
		[] -> check_disabled_functions(TailA,B);
		Other -> [make_pair(Func,Other)|check_disabled_functions(TailA,B)]
	end.

%% The input parameters are two lists, and the output is the 
%% intersection of the two lists.
lists_intersection([],_)->[];
lists_intersection(_,[])->[];
lists_intersection([HeadA|TailA],[HeadB|TailB])->
	if
		HeadA<HeadB -> lists_intersection(TailA,[HeadB|TailB]);
		HeadA>HeadB -> lists_intersection([HeadA|TailA],TailB);
		HeadA==HeadB -> [HeadA|lists_intersection(TailA,TailB)]
	end.
